Require Export SetChecker.
Require Export NArith.

Section Pythagoras.

Definition pythagorean (a b c:positive) := (a*a + b*b = c*c)%positive.

Lemma pythagorean_comm : forall a b c, pythagorean a b c -> pythagorean b a c.
unfold pythagorean; intros.
rewrite Pos.add_comm; auto.
Qed.

Lemma pythagorean_neq_2 : forall a b c, pythagorean a b c -> b <> c.
unfold pythagorean; red; intros.
rewrite H0 in H.
revert H; apply Pos.add_no_neutral.
Qed.

Lemma pythagorean_neq_1 : forall a b c, pythagorean a b c -> a <> c.
red; intros.
apply (pythagorean_neq_2 b a c); auto.
apply pythagorean_comm; auto.
Qed.

Definition coloring := positive -> bool.

Definition pythagorean_pos (C:coloring) :=
  forall a b c, pythagorean a b c -> (C a <> C b) \/ (C a <> C c) \/ (C b <> C c).

Definition pythagorean_neg (C:coloring) :=
  forall a b c, C a = C b -> C a = C c -> ~pythagorean a b c.

Theorem pythagorean_equiv : forall C, pythagorean_pos C <-> pythagorean_neg C.
unfold pythagorean_pos, pythagorean_neg; split; intros.
intro.
elim (H _ _ _ H2); intros; auto.
inversion_clear H3; auto.
apply H4; transitivity (C a); auto.

generalize (H a b c).
elim (C a); elim (C b); elim (C c); auto; intros.
elim H1; auto.
right; right; discriminate.
left; discriminate.
elim H1; auto.
Qed.

Definition pyth_lim_pos (n:positive) (C:coloring) :=
  forall a b c, (a<n)%positive -> (b<n)%positive -> (c<n)%positive -> pythagorean a b c -> (C a <> C b) \/ (C a <> C c) \/ (C b <> C c).

Definition pyth_lim_neg (n:positive) (C:coloring) :=
  forall a b c, (a<n)%positive -> (b<n)%positive -> (c<n)%positive -> C a = C b -> C a = C c -> ~pythagorean a b c.

Theorem pythagorean_lim_equiv : forall n C, pyth_lim_pos n C <-> pyth_lim_neg n C.
unfold pyth_lim_pos, pyth_lim_neg; split; intros.
intro.
elim (H _ _ _ H0 H1 H2 H5); intros; auto.
inversion_clear H6; auto.
apply H7; transitivity (C a); auto.

generalize (H a b c H0 H1 H2).
elim (C a); elim (C b); elim (C c); auto; intros.
elim H4; auto.
right; right; discriminate.
left; discriminate.
elim H4; auto.
Qed.

Lemma pyth_lim_gen : forall n C, pythagorean_pos C -> pyth_lim_pos n C.
intros.
red; red in H; intros; auto.
Qed.

Lemma pyth_lim_gen_neg : forall n C, ~pyth_lim_neg n C -> ~pythagorean_neg C.
intros; intro.
apply H.
rewrite <- pythagorean_lim_equiv.
apply pyth_lim_gen.
rewrite pythagorean_equiv; auto.
Qed.

Theorem no_pythagorean : forall n, (forall C, ~pyth_lim_pos n C) ->
                         forall C, ~pythagorean_pos C.
intros; intro.
elim (H C).
apply pyth_lim_gen; auto.
Qed.

Lemma pythagorean_dec : forall a b c, {pythagorean a b c} + {~pythagorean a b c}.
intros; unfold pythagorean.
elim (Pos.eq_dec (a*a+b*b) (c*c)); auto.
Defined.

End Pythagoras.

Section Lists_and_Stuff.

Definition target_list := list (list positive).

Definition ok_list (t:target_list) := forall tuple,
  In tuple t -> exists a b c, tuple = (a::b::c::nil) /\ pythagorean a b c.

Definition colorful (C:coloring) (t:target_list) := forall tuple,
  In tuple t -> exists a b, In a tuple /\ In b tuple /\ C a <> C b.

Fixpoint inner_cycle (n:positive) (m:nat) (b:positive) : target_list :=
  match m with
  | 0 => nil
  | S m' => let mN := Pos.of_nat m in let sqrt := (Pos.sqrt (n*n+mN*mN)) in
            if (sqrt<?b)%positive 
            then if pythagorean_dec n mN sqrt
                 then (mN::n::sqrt::nil) :: inner_cycle n m' b
                 else (inner_cycle n m' b)
            else (inner_cycle n m' b)
  end.

Lemma inner_cycle_okl : forall n m b, ok_list (inner_cycle n m b).
induction m; red; simpl.
intros; inversion H.
intros b tuple; set (mN := Pos.of_nat (S m)); simpl in mN; fold mN.
elim (Pos.sqrt (n*n + mN*mN) <? b)%positive.
2: apply (IHm b); auto.
set (Z := (pythagorean_dec n mN (Pos.sqrt (n*n + mN*mN)))).
assert (Z = (pythagorean_dec n mN (Pos.sqrt (n*n + mN*mN)))); auto; induction Z.
2: apply (IHm b); auto.
intro; inversion_clear H0.
2: apply (IHm b); auto.
exists mN; exists n; exists (Pos.sqrt (n*n + mN*mN)); split; auto.
red; red in a; rewrite <- a; apply Pos.add_comm.
Qed.

Fixpoint outer_cycle (n:nat) (b:positive) : target_list :=
  match n with
  | 0 => nil
  | S m => (inner_cycle (Pos.of_nat n) n b) ++ (outer_cycle m b)
  end.

Lemma outer_cycle_okl : forall n b, ok_list (outer_cycle n b).
induction n; red; intros.
inversion H.
elim (in_app_or _ _ _ H); intros.
eapply inner_cycle_okl; eauto.
apply (IHn b); auto.
Qed.

Definition BPT_list (n:nat) := outer_cycle n (Pos.of_nat n).

End Lists_and_Stuff.

Section Lists_of_Clauses.

Definition posF (tuple:list positive) : Clause := map pos tuple.
Definition negF (tuple:list positive) : Clause := map neg tuple.

Fixpoint target_formula (t:target_list) : CNF :=
  match t with
  | nil => nil
  | tuple::t' => (posF tuple) :: (negF tuple) :: (target_formula t')
  end.

Lemma pos_in : forall a t, In a t -> In (pos a) (posF t).
intros; apply in_map; auto.
Qed.

Lemma neg_in : forall a t, In a t -> In (neg a) (negF t).
intros; apply in_map; auto.
Qed.

Lemma target_formula_in : forall a cl t, In (pos a) cl \/ In (neg a) cl -> 
  In cl (target_formula t) -> exists l, In a l /\ In l t.
intros a cl t; revert a cl.
induction t; simpl; intros.
inversion H0.
inversion_clear H0.
(* 1/3 *)
rewrite <- H1 in H; inversion_clear H.
exists a; split; auto.
elim (map_in _ _ _ _ _ H0); intros.
inversion_clear H; inversion H2; rewrite <- H4; auto.
elim (map_in _ _ _ _ _ H0); intros.
inversion_clear H; inversion H2.

inversion_clear H1.
(* 2/3 *)
rewrite <- H0 in H; inversion_clear H.
elim (map_in _ _ _ _ _ H1); intros.
inversion_clear H; inversion H2.
exists a; split; auto.
elim (map_in _ _ _ _ _ H1); intros.
inversion_clear H; inversion H2; rewrite <- H4; auto.
(* 3/3 *)
elim (IHt a0 cl); intros; auto.
inversion_clear H1; exists x; split; auto.
Qed.

Definition BPT_formula (n:nat) := target_formula (BPT_list n).

Lemma colorful_gives_val : forall C t,
  colorful C t -> satisfies C (target_formula t).
intros; apply for_all_satisfies.
induction t; simpl; intros.
inversion H0.
rename a into tuple.
elim (H tuple); simpl; auto; intros a Ha; inversion_clear Ha.
rename x into b; inversion_clear H1; inversion_clear H3; inversion_clear H0.

set (Z := C a); assert (Z = C a); auto; induction Z.
(* 1/5 *)
rewrite <- H3; apply exists_C_satisfies with (pos a).
apply pos_in; auto.
simpl; rewrite <- H0; auto.
(* 2/5 *)
rewrite <- H3; apply exists_C_satisfies with (pos b); auto.
apply pos_in; auto.
simpl; replace (C b) with true; auto.
symmetry; apply not_false_is_true.
rewrite H0; auto.

inversion_clear H3.
set (Z := C a); assert (Z = C a); auto; induction Z.
(* 3/5 *)
rewrite <- H0; apply exists_C_satisfies with (neg b); auto.
apply neg_in; auto.
simpl; replace (C b) with false; auto.
symmetry; apply not_true_is_false.
rewrite H3; auto.
(* 4/5 *)
rewrite <- H0; apply exists_C_satisfies with (neg a).
apply neg_in; auto.
simpl; rewrite <- H3; auto.

(* 5/5 *)
apply IHt; auto.
red; intros; apply H; simpl; auto.
Qed.

Lemma val_gives_colorful : forall C t, satisfies C (target_formula t) ->
  colorful C t.
induction t; red; intros; inversion_clear H0; rename a into l.

rewrite H1 in H; clear H1 l.
assert (C_satisfies C (posF tuple)).
apply H; simpl; auto.
elim (C_satisfies_exists _ _ H0); clear H0; intros.
inversion_clear H0.
elim (map_in _ _ _ _ _ H1); clear H1; intros.
inversion_clear H0.
rename x0 into a; rewrite <- H1 in H2; clear H1 x.

assert (C_satisfies C (negF tuple)).
apply H; simpl; auto.
elim (C_satisfies_exists _ _ H0); clear H0; intros.
inversion_clear H0.
elim (map_in _ _ _ _ _ H1); clear H1; intros.
inversion_clear H0.
rename x0 into b; rewrite <- H1 in H4; clear H1 x.

exists a; exists b; repeat split; auto.
intro.
simpl in H2; simpl in H4; rewrite H0 in H2.
induction (C b); [inversion H4 | inversion H2].

apply IHt; auto.
simpl in H.
red; intros; apply H; simpl; auto.
Qed.

Definition reduces (C C':CNF) := forall V, satisfies V C -> exists V', satisfies V' C'.

End Lists_of_Clauses.

Section Formula.

Parameter TheN : nat.

Definition The_CNF := BPT_formula TheN.

End Formula.

Lemma BPT_formula_sat : forall (C:coloring), pythagorean_pos C ->
  forall n, satisfies C (BPT_formula n).
simpl; intros C HC; unfold BPT_formula, BPT_list.
(* we abstract from the bound *)
assert (forall n b, satisfies C (target_formula (outer_cycle n b))); auto.
intros; apply colorful_gives_val.
red; intros.
elim (outer_cycle_okl _ _ _ H); intros.
rename b into b'; rename x into a; inversion_clear H0.
rename x into b; inversion_clear H1.
rename x into c; inversion_clear H0.
elim (HC _ _ _ H2); clear HC; intros.
rewrite H1; exists a; exists b; repeat split; simpl; auto.
inversion_clear H0.
rewrite H1; exists a; exists c; repeat split; simpl; auto.
rewrite H1; exists c; exists b; repeat split; simpl; auto.
Qed.

Theorem Pythagorean_Theorem : unsat The_CNF -> forall C, ~pythagorean_pos C.
intros; intro.
apply (H C).
apply BPT_formula_sat; auto.
Qed.

Section Simplification.

Definition change_val (p:positive) (b:bool) (V:Valuation) :=
  fun p' => if (p=?p')%positive then b else V p'.

Lemma change_val_p : forall p b V, change_val p b V p = b.
intros.
unfold change_val.
replace (p=?p)%positive with true; auto.
symmetry; rewrite Pos.eqb_eq; auto.
Qed.

Lemma change_val_p' : forall p b V p', p <> p' -> change_val p b V p' = V p'.
intros.
unfold change_val.
replace (p=?p')%positive with false; auto.
elim (Pos.eqb_eq p p'); intros.
symmetry; apply not_true_is_false; intro; apply H; auto.
Qed.

Fixpoint no_occurrence (p:positive) (t:target_list) :=
  match t with
  | nil => True
  | (l::t') => ~In p l /\ no_occurrence p t'
  end.

Lemma no_occurrence_char : forall p t,
  no_occurrence p t <-> forall l, In l t -> ~In p l.
induction t; split; simpl; auto.
intros.
inversion_clear H.
inversion_clear H0; auto.
rewrite <- H; auto.
apply IHt; auto.
split; auto.
apply IHt; auto.
Qed.

Fixpoint one_occurrence (p:positive) (t:target_list) :=
  match t with
  | nil => False
  | (l::t') => (In p l /\ no_occurrence p t')
                \/ (~In p l /\ one_occurrence p t')
  end.

Lemma one_occurrence_find : forall p t, one_occurrence p t ->
  exists l, In p l /\ In l t /\ (forall l', In l' t -> l <> l' -> ~In p l').
induction t; simpl; intros; inversion_clear H; inversion_clear H0.
exists a; repeat split; auto.
intros; intro.
elim (no_occurrence_char p t); intros.
apply H4 with l'; auto.
inversion_clear H0; auto.
elim H2; auto.

elim (IHt H1); intros.
inversion_clear H0; inversion_clear H3; exists x; repeat split; auto.
intros.
inversion_clear H3; auto.
rewrite <- H6; auto.
Qed.

Lemma colorful_add : forall t a, no_occurrence a t ->
  forall C, colorful C t -> forall b, a <> b ->
  forall l, In a l -> In b l -> exists C', colorful C' (l::t).
intros.
exists (change_val a (negb (C b)) C); red; intros.
inversion_clear H4.
rewrite <- H5; exists a; exists b; repeat split; simpl; auto.
rewrite change_val_p; rewrite change_val_p'; auto.
apply no_fixpoint_negb.

clear H3 H2 l H1.
elim H0 with tuple; auto.
intros a' H'; inversion_clear H'.
rename x into b'; inversion_clear H1.
inversion_clear H3.
exists a'; exists b'; repeat split; auto.
elim (no_occurrence_char a t); intros.
generalize (H3 H); clear H6 H3 H; intros.
repeat rewrite change_val_p'; auto.
intro; revert H1; rewrite <- H3; apply H; auto.
intro; revert H2; rewrite <- H3; apply H; auto.
Qed.

Lemma no_occurrence_add : forall t a, no_occurrence a t ->
  forall b c, a <> b -> a <> c ->
  reduces (target_formula t) (target_formula ((a::b::c::nil)::t)).
red; intros.
generalize (val_gives_colorful _ _ H2); intros.
elim (colorful_add _ _ H _ H3 b H0 (a::b::c::nil)); intros.
2: simpl; auto.
2: simpl; auto.
rename x into C; exists C.
apply colorful_gives_val; auto.
Qed.

Fixpoint remove_number (p:positive) (t:target_list) :=
  match t with
  | nil => nil
  | l::t' => if In_dec Pos.eq_dec p l then remove_number p t' else l::remove_number p t'
  end.

Lemma remove_no_occurrence : forall p t, no_occurrence p (remove_number p t).
induction t; simpl; auto.
elim in_dec; simpl; auto.
Qed.

Lemma remove_keep : forall p t l, In l t -> ~In p l -> In l (remove_number p t).
induction t; simpl; auto.
intros; elim in_dec; simpl; auto.
intros; inversion_clear H; auto.
elim H0; rewrite <- H1; auto.
intros; inversion_clear H; auto.
Qed.

Lemma remove_in : forall p t l, In l (remove_number p t) -> In l t.
induction t; simpl; auto.
elim in_dec; simpl; auto.
intros; inversion_clear H; auto.
Qed.

Lemma colorful_remove_singleton : forall t a, one_occurrence a t -> ok_list t ->
  forall C, colorful C (remove_number a t) ->
  exists C', colorful C' t.
intros.

elim (one_occurrence_find _ _ H); intros l Hl.
elim Hl; clear Hl; intros Hal Hlt.
elim Hlt; clear Hlt; intros Hlt Hl.
elim (H0 l); clear H0; intros; auto.
rename x into a'; inversion_clear H0.
rename x into b'; inversion_clear H2.
rename x into c'; inversion_clear H0.

elim (Pos.eq_dec a c'); intro.
(* case 1 *)
elim (colorful_add (remove_number a t) a) with C a' l; intros; auto.
2: apply remove_no_occurrence.
3: rewrite H2; simpl; auto.
2: intro; apply pythagorean_neq_1 with a' b' c'; auto; rewrite <- a0; auto.
rename x into C'; exists C'.
red; intros.
apply H0.
elim (list_eq_dec Pos.eq_dec tuple l); intros; simpl; auto.
right; apply remove_keep; auto.

(* case 2 *)
elim (colorful_add (remove_number a t) a) with C c' l; intros; auto.
2: apply remove_no_occurrence.
2: rewrite H2; simpl; auto.
rename x into C'; exists C'.
red; intros.
apply H0.
elim (list_eq_dec Pos.eq_dec tuple l); intros; simpl; auto.
right; apply remove_keep; auto.
Qed.

Lemma no_occurrence_dec : forall p t, {no_occurrence p t}+{~no_occurrence p t}.
induction t.
left; simpl; auto.
elim (in_dec Pos.eq_dec p a); intro.
right; intro.
inversion_clear H; auto.
elim IHt; intro.
left; split; auto.
right; intro.
inversion_clear H; auto.
Qed.

Lemma one_occurrence_dec : forall p t, {one_occurrence p t}+{~one_occurrence p t}.
induction t.
right; simpl; auto.
elim (in_dec Pos.eq_dec p a); intro.
elim (no_occurrence_dec p t); intro.
left; left; split; auto.
right; intro; inversion_clear H; inversion_clear H0; auto.
inversion_clear IHt.
left; right; split; auto.
right; intro; inversion_clear H0; inversion_clear H1; auto.
Qed.

Fixpoint simplify (t:target_list) (l:list positive) :=
  match l with
  | nil => t
  | p::l' => if (one_occurrence_dec p t) then simplify (remove_number p t) l'
             else simplify t l'
  end.

Lemma in_simplify : forall l l' t, In l' (simplify t l) -> In l' t.
induction l; simpl; auto.
intros l' t; elim one_occurrence_dec; simpl; auto.
intros; apply remove_in with a; auto.
Qed.

Lemma colorful_simplify : forall t, ok_list t ->
  forall l C, colorful C (simplify t l) -> exists C', colorful C' t.
intros.
revert t H C H0.
induction l; simpl; intros.
exists C; auto.
revert H0; elim one_occurrence_dec; intros.
elim (IHl (remove_number a t)) with C; auto; intros.
apply colorful_remove_singleton with a x; auto.
red; intros; apply H.
apply remove_in with a; auto.
apply IHl with C; auto.
Qed.

Definition simplified_BPT_formula (n:nat) (l:list positive) :=
  target_formula (simplify (BPT_list n) l).

Parameter The_List : list positive.

Definition The_Simple_CNF := simplified_BPT_formula TheN The_List.

(* The stronger result *)

Theorem simplification_ok : unsat The_CNF <-> unsat The_Simple_CNF.
unfold The_CNF, The_Simple_CNF, BPT_formula, simplified_BPT_formula.
split; repeat intro.

generalize (val_gives_colorful _ _ H0); intros.
elim (colorful_simplify (BPT_list TheN)) with The_List v; auto; intros.
2: apply outer_cycle_okl.
rename x into C; apply H with C.
apply colorful_gives_val; auto.

apply (H v).
apply colorful_gives_val.
generalize (val_gives_colorful _ _ H0); intro.
red; intros; apply H1.
apply in_simplify with The_List; auto.
Qed.

Theorem Pythagorean_Theorem' : unsat The_Simple_CNF -> forall C, ~pythagorean_pos C.
intros; intro.
apply Pythagorean_Theorem with C.
2: auto.
rewrite simplification_ok; auto.
Qed.

End Simplification.

Section SymmetryBreak.

Lemma fix_one_color : forall C, pythagorean_pos C ->
  forall n b, exists C', pythagorean_pos C' /\ C' n = b.
intros.
elim (bool_dec b (C n)); intros.
exists C; auto.
exists (fun n => negb (C n)); split; auto.
red; intros.
elim (H _ _ _ H0); intros.
left; intro; apply H1.
rewrite <- (negb_involutive (C a)); rewrite <- (negb_involutive (C b1)); rewrite H2; auto.
inversion_clear H1.
right; left; intro; apply H2.
rewrite <- (negb_involutive (C a)); rewrite <- (negb_involutive (C c)); rewrite H1; auto.
right; right; intro; apply H2.
rewrite <- (negb_involutive (C b1)); rewrite <- (negb_involutive (C c)); rewrite H1; auto.

induction b.
rewrite (not_true_is_false (C n)); auto.
rewrite (not_false_is_true (C n)); auto.
Qed.

Parameter TheBreak : positive.

Definition The_Asymmetric_CNF : CNF :=
  (pos TheBreak :: nil) :: simplified_BPT_formula TheN The_List.

(* The stronger result *)

Theorem symbreak_ok : unsat The_CNF <-> unsat The_Asymmetric_CNF.
rewrite simplification_ok.
unfold The_Simple_CNF, The_Asymmetric_CNF, simplified_BPT_formula.
split; repeat intro.

apply (H v).
red; intros; apply H0; simpl; auto.

elim (bool_dec true (v TheBreak)); intros.
apply (H v); auto.
generalize (satisfies_for_all _ _ H0); clear H0; intros.
apply for_all_satisfies; intros.
inversion_clear H1; auto.
rewrite <- H2; simpl; rewrite <- a; auto.

set (v' := fun n => negb (v n)).
apply (H v'); auto.
generalize (satisfies_for_all _ _ H0); clear H0; intros.
apply for_all_satisfies; red; intros.
inversion_clear H1; auto.
rewrite <- H2; simpl.
replace (v' TheBreak) with true; auto.
unfold v'; clear v'.
symmetry; apply not_false_is_true; intro.
apply b; rewrite <- (negb_involutive (v TheBreak)); rewrite H1; auto.

clear H; revert H0 H2.
unfold simplified_BPT_formula.
induction (simplify (BPT_list TheN) The_List); simpl; intros; auto.
inversion_clear H2.
inversion_clear H2.

assert (C_satisfies v (negF a)).
apply H0; simpl; auto.
elim (C_satisfies_exists _ _ H1); intros.
inversion_clear H2.
apply exists_C_satisfies with (negate x).
rewrite <- H; clear H H4 H1 H0 IHt t; induction a.
inversion H3.
inversion_clear H3; simpl.
left; rewrite <- H; auto.
right; apply IHa; auto.
unfold v'.
revert H4; clear H3 H1 H H0 IHt t a v'; induction x; simpl; auto; elim (v p); simpl; auto.

inversion_clear H.

assert (C_satisfies v (posF a)).
apply H0; simpl; auto.
elim (C_satisfies_exists _ _ H); intros.
inversion_clear H2.
apply exists_C_satisfies with (negate x).
rewrite <- H1; clear H H4 H1 H0 IHt t; induction a.
inversion H3.
inversion_clear H3; simpl.
left; rewrite <- H; auto.
right; apply IHa; auto.
unfold v'.
revert H4; clear H3 H1 H H0 IHt t a v'; induction x; simpl; auto; elim (v p); simpl; auto.

apply IHt; auto.
Qed.

Theorem Pythagorean_Theorem'' : unsat The_Asymmetric_CNF -> forall C, ~pythagorean_pos C.
intros; intro.
apply Pythagorean_Theorem with C; auto.
rewrite symbreak_ok; auto.
Qed.

End SymmetryBreak.

